# frozen_string_literal: true

require_relative 'xml_parsable'

#
module DwCR
  # This module provides functionality to create, update, and verify the
  # part of the DwCR schema needed to persist information about the DwCR
  module Metaschema
    # Creates the tables for the metaschema present in every DwCR
    # _archives_, _entities_, _attributes_, _content_files_
    # loads the Sequel::Model classes for these tables
    def self.create
      tabledefs = Psych.load_file(File.join(__dir__, 'metaschema_tables.yml'))
      tabledefs.to_h.each do |table, columns|
        DB.create_table? table do
          primary_key :id
          columns.each { |c| column(*c) }
        end
      end
      load_models
    end

    # Loads all Sequel::Model classes for the metaschema
    def self.load_models
      require_relative 'archive'
      require_relative 'entity'
      require_relative 'attribute'
      require_relative 'content_file'
    end

    # Returns schema or index parameters for a table, depending on the
    # second argument (+:schema+ or +:indexes+)
    # returns +false+ if the table does not exist
    def self.inspect_table(table, method)
      DB.indexes(table).values.map { |x| x[:columns] }.flatten
      DB.send(method, table)
    rescue Sequel::Error
      false
    end

    # Performs an integrety check on +table+, veryfies all columns
    # are present with the parameters given in +columns+;
    # a column parameter is an array with the structure:
    # <tt>[:column_name, :column_type, {column_options} ]</tt>
    def self.columns?(table, *columns)
      db_cols = inspect_table(table, :schema)
      return unless db_cols
      exp_cols = columns.map(&:first).unshift(:id)
      exp_cols == db_cols.map(&:first)
    end

    # Performs an integrety check on +table+, veryfies all indices
    # are present with the parameters given in +columns+;
    # a column parameter is an array with the structure:
    # <tt>[:column_name, :column_type, {column_options} ]</tt>
    def self.indexes?(table, *columns)
      db_idxs = inspect_table(table, :indexes)
      return unless db_idxs
      exp_idxs = columns.select { |column| column[2]&.fetch(:index, false) }
                        .map(&:first)
      exp_idxs & db_idxs.values.map { |x| x[:columns] }.flatten == exp_idxs
    end

    # Updates all Attribute instances in a Archive
    # with parameters from files in ContentFile
    # _schema_options_: a Hash with attribute names as keys and boolean values
    # <tt>{ :type => true, :length => true }</tt>
    # updates any attribute given as key where value is _true_
    def self.update(archive, **options)
      return if options.empty?

      # FIXME: handle situation where schema tables have been created
      options.select! { |_k, v| v == true }
      archive.entities
             .each { |entity| entity.update_attributes!(*options.keys) }
    end

    # Performs an integrity check on the metaschema in the DWCR file
    # (the current database connection)
    # returns true if all tables, columns, and indices as given in
    # _config/metaschema_tables.yml_ are present
    def self.valid?
      tabledefs = Psych.load_file('lib/dwcr/metaschema/metaschema_tables.yml')
      status = tabledefs.map do |td|
        table = td.first
        columns = td.last
        columns?(table, *columns) && indexes?(table, *columns)
      end
      return false if status.uniq.size > 1
      status.first
    end
  end
end
